@@ -1,15 +1,25 @@
 /*
  * PAM module for kerberized coda only.
- * This module should be used as "optional" in "session", NOT
- * for authentication.
+ * This module should be used as "optional" in both "auth" and "session",
+ * but NOT for authentication,
+ * it just acquires a Coda token if there is a Kerberos one.
+ *
+ * ver 0.3, Nov 2001
+ *  moved token acuisition from open_session() to setcred()
+ *  to suit existing pam framework for credentials setting
+ * arguments (a minor change):
+ *  auth: ignore_root, kclog /path/to/kclog/binary
+ *  session: nocunlog, cunlog /path/to/cunlog/binary
+ *
  * ver 0.2
  *  aaargh! memory allocation errors fixed
+ *
  * ver 0.1
  * arguments:
- *  open_sesssion: ignore_root, clog /path/to/kclog/binary
+ *  open_session: ignore_root, clog /path/to/kclog/binary
  *  close_session: nocunlog, cunlog /path/to/cunlog/binary
  *
- * it runs kclog program at open_session() to aquire Coda tokens
+ * it runs kclog program at setcred() to aquire Coda tokens
  * and optionally cunlog at close_session() to destroy them.
  *
  * use kerberos pam module for authentication, then the things will work
@@ -62,8 +72,15 @@
 #include <security/_pam_macros.h>
 
 PAM_EXTERN
-int pam_sm_open_session(pam_handle_t * pamh, int flags, int argc
-			,const char **argv)
+int pam_sm_authenticate(pam_handle_t * pamh, int flags,
+                        int argc, const char **argv)
+{
+  return PAM_IGNORE;
+}
+
+PAM_EXTERN
+int pam_sm_setcred(pam_handle_t * pamh, int flags,
+                   int argc, const char **argv)
 {
     int retval;
     const char *user = NULL;
@@ -73,32 +90,35 @@ int pam_sm_open_session(pam_handle_t * pamh, int flags, int argc
     pid_t code;
     int status, number = 0, pipedes[2];
     char *cargv[16];
-    char *clog_program;
+    char *kclog_program;
     char **en;
 
-    clog_program = malloc ((strlen(CLOG_PROGRAM)+1)*sizeof(char));
-    strcpy (clog_program,CLOG_PROGRAM);
+    if (flags != PAM_ESTABLISH_CRED)
+      return PAM_CRED_ERR;
+
+    kclog_program = malloc ((strlen(CLOG_PROGRAM)+1)*sizeof(char));
+    strcpy (kclog_program,CLOG_PROGRAM);
 
 /* Parse parameters for module */
     for (; argc-- > 0; argv++) {
 	if (strcmp(*argv, "ignore_root") == 0)
 	    ignore_root = 1;
-        else if (strcmp(*argv,"clog") == 0) {
+        else if (strcmp(*argv,"kclog") == 0) {
             if (argc-- > 0) {
-            clog_program = realloc (clog_program,(strlen(*(++argv))+1)*sizeof(char));
-            strcpy (clog_program,*argv);
+            kclog_program = realloc (kclog_program,(strlen(*(++argv))+1)*sizeof(char));
+            strcpy (kclog_program,*argv);
             }
 	}
     }
 
     retval = pam_get_user(pamh, &user, NULL);
     if (retval != PAM_SUCCESS)
-        return PAM_SESSION_ERR;
+        return PAM_USER_UNKNOWN;
 
     if (user == NULL || !isalnum(*user)) {
         if (user)
             syslog(LOG_ERR, "pam_kcoda: bad username [%s]", user);
-        return PAM_SESSION_ERR;
+        return PAM_USER_UNKNOWN;
     }
 
 /* Get User Information */
@@ -106,38 +126,38 @@ int pam_sm_open_session(pam_handle_t * pamh, int flags, int argc
 
 /* User does not exist */
     if (pw == NULL)
-        return PAM_SESSION_ERR;
+        return PAM_USER_UNKNOWN;
 
 /* Handle Root */
     if (pw->pw_uid == 0 && ignore_root == 1)
 	return PAM_SUCCESS;
 
-/* set up pipe for communication with clog */
+/* set up pipe for communication with kclog */
     if (pipe(pipedes) != 0) {
 	syslog(LOG_CRIT, "pam_kcoda: Can't open pipe");
-	return PAM_SESSION_ERR;
+	return PAM_CRED_ERR;
     }
 
-/* set up arguments for clog */
-    cargv[number++] = "clog";
+/* set up arguments for kclog */
+    cargv[number++] = "kclog";
     cargv[number++] = user;
     cargv[number] = NULL;
 
     code = fork();
     if (code == -1) {
 	syslog(LOG_ERR, "pam_kcoda: fork faild");
-	return PAM_SESSION_ERR;
+	return PAM_CRED_ERR;
     }
     if (code == 0) {
 	if (setgid(pw->pw_gid) < 0) {
 	    syslog(LOG_ERR, "pam_kcoda: setgid(%d) failed",
 		   pw->pw_gid);
-	    return PAM_SESSION_ERR;
+	    return PAM_CRED_ERR;
 	}
 	if (setuid(pw->pw_uid) < 0) {
 	    syslog(LOG_ERR, "pam_kcoda: setuid(%d) failed",
 		   pw->pw_uid);
-	    return PAM_SESSION_ERR;
+	    return PAM_CRED_ERR;
 	}
 	close(0);		/* stdin */
 	dup(pipedes[0]);
@@ -148,9 +168,9 @@ int pam_sm_open_session(pam_handle_t * pamh, int flags, int argc
 	close(2);
 	dup(1);
         if( (en=pam_getenvlist(pamh)) != NULL )
-          execve(clog_program, cargv, en);
+          execve(kclog_program, cargv, en);
         else
-          execv(clog_program, cargv);
+          execv(kclog_program, cargv);
 	/* close pipe */
 	close(0);		/* stdin */
 	close(1);		/* stdout */
@@ -169,19 +189,26 @@ int pam_sm_open_session(pam_handle_t * pamh, int flags, int argc
     if( code == -1 )
 /* we did't get the child status (why?)
    or got interrupt - it might be ok but we are paranoid :-) */
-	return PAM_SESSION_ERR;
+	return PAM_CRED_ERR;
     user = NULL;		/* clean up */
     if (WIFEXITED(status))
 	if (WEXITSTATUS(status) == 0)
 	    return PAM_SUCCESS;
-    syslog(LOG_NOTICE, "pam_kcoda: pam_clog[%d]: unsuccessful",
+    syslog(LOG_NOTICE, "pam_kcoda: pam_kclog[%d]: unsuccessful",
 	   getpid());
-    return PAM_SESSION_ERR;
+    return PAM_CRED_ERR;
 }
 
 PAM_EXTERN
-int pam_sm_close_session(pam_handle_t * pamh, int flags, int argc
-			,const char **argv)
+int pam_sm_open_session(pam_handle_t * pamh, int flags,
+                        int argc ,const char **argv)
+{
+  return PAM_SUCCESS;
+}
+
+PAM_EXTERN
+int pam_sm_close_session(pam_handle_t * pamh, int flags,
+                         int argc ,const char **argv)
 {
     int retval;
     const char *user = NULL;
